package de.fhhannover.inform.trust.ifmapcli;

import java.io.FileNotFoundException;
import java.io.InputStream;

import javax.net.ssl.TrustManager;

import org.w3c.dom.Document;

import de.fhhannover.inform.trust.ifmapcli.common.Common;
import de.fhhannover.inform.trust.ifmapcli.common.Config;
import de.fhhannover.inform.trust.ifmapj.IfmapJ;
import de.fhhannover.inform.trust.ifmapj.IfmapJHelper;
import de.fhhannover.inform.trust.ifmapj.binding.IfmapStrings;
import de.fhhannover.inform.trust.ifmapj.channel.SSRC;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapErrorResult;
import de.fhhannover.inform.trust.ifmapj.exception.IfmapException;
import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;
import de.fhhannover.inform.trust.ifmapj.identifier.Identifier;
import de.fhhannover.inform.trust.ifmapj.identifier.Identifiers;
import de.fhhannover.inform.trust.ifmapj.messages.MetadataLifetime;
import de.fhhannover.inform.trust.ifmapj.messages.PublishDelete;
import de.fhhannover.inform.trust.ifmapj.messages.PublishRequest;
import de.fhhannover.inform.trust.ifmapj.messages.PublishUpdate;
import de.fhhannover.inform.trust.ifmapj.messages.Requests;
import de.fhhannover.inform.trust.ifmapj.metadata.StandardIfmapMetadataFactory;

/**
 * A simple tool that publishes or deletes discovered-by metadata on a link<br/>
 * between ip and device identifiers.<br/>
 * When metadata is published, the lifetime is set to be 'forever'.
 * 
 * @author ib
 *
 */
public class IpDiscBy {
	final static String CMD = "ip-disc-by";
	final static int MIN_ARGS = 3;			// update|delete, ip, dev
	final static int EXPECTED_ARGS = 8;		// update|delete, ip, dev
											// url, user, pass,
											// keystorePath, keystorePass
	
	// in order to create the necessary objects, make use of the appropriate
	// factory classes
	private static StandardIfmapMetadataFactory mf = IfmapJ
			.createStandardMetadataFactory();
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		String op, dev, ip;
		Config cfg;
		SSRC ssrc;
		PublishRequest req;
		PublishUpdate publishUpdate;
		PublishDelete publishDelete;
		TrustManager[] tms;
		Identifier devIdentifier;
		Identifier ipIdentifier;
		Document metadata;
		InputStream is;
		
		// check number of mandatory command line arguments
		if(args.length < MIN_ARGS){
			IpDiscBy.usage();
			return;
		}
		
		// parse mandatory command line arguments
		op = args[0];
		ip = args[1];
		dev = args[2];
		if(Common.isUpdateorDelete(op) == false){
			IpDiscBy.usage();
			return;
		}
		
		// check and load optional parameters
		cfg = Common.checkAndLoadParams(args, EXPECTED_ARGS);
		System.out.println(CMD + " uses config " + cfg);
		
		// prepare identifiers
		ipIdentifier = Identifiers.createIp4(ip);
		devIdentifier = Identifiers.createDev(dev);
		
		// prepare metadata
		metadata = mf.createDiscoveredBy();
		
		// update or delete
		if(Common.isUpdate(op)){
			publishUpdate = Requests.createPublishUpdate(ipIdentifier, devIdentifier,
					metadata, MetadataLifetime.forever);			
			req = Requests.createPublishReq(publishUpdate);
		} else {
			String filter = "meta:discovered-by";
			publishDelete = Requests.createPublishDelete(ipIdentifier, devIdentifier, filter);
			publishDelete.addNamespaceDeclaration(IfmapStrings.STD_METADATA_PREFIX,
					IfmapStrings.STD_METADATA_NS_URI);
			req = Requests.createPublishReq(publishDelete);
		}
				
		// publish
		try {
			is = Common.prepareTruststoreIs(cfg.getTruststorePath());
			tms = IfmapJHelper.getTrustManagers(is, cfg.getTruststorePass());
			ssrc = IfmapJ.createSSRC(cfg.getUrl(), cfg.getUser(), cfg.getPass(), tms);
			ssrc.newSession();
			ssrc.publish(req);
			ssrc.endSession();
		} catch (InitializationException e) {
			System.out.println(e.getDescription() + " " + e.getMessage());
		} catch (IfmapErrorResult e) {
			System.out.println(e.getErrorString());
		} catch (IfmapException e) {
			System.out.println(e.getDescription() + " " + e.getMessage());
		} catch (FileNotFoundException e) {
			System.out.println(e.getMessage());
		}
	}
	
	private static void usage() {
		System.out.println("usage:\n" +
				"\t" + IpDiscBy.CMD + " update|delete ip dev " +
				"[url user pass truststore truststorePass]");
		System.out.println(Common.USAGE);
	}
}
